# -*- coding: utf-8 -*-

import zlib
import logging
from contextlib import contextmanager

import redis
from redis.client import StrictRedis, StrictPipeline


class TracingPipeline(StrictPipeline):
    """Simple sublcass that logs pipeline stacks to the trace logger
    """
    def __init__(self, *args, **kwargs):
        self.trace = kwargs.pop('trace', False)
        super(TracingPipeline, self).__init__(*args, **kwargs)

    def execute(self, *args, **kwargs):
        log = logging.getLogger('redis.trace.pipeline')
        # self.connection_pool.total_commands += 1
        for num, cmd in enumerate(self.command_stack):
            self.connection_pool.total_pipeline_commands += 1
            if self.trace:
                log.debug('[%04d:%03d] %s',
                          self.connection_pool.total_commands, num, cmd[0])
        return StrictPipeline.execute(self, *args, **kwargs)


class TracingRedis(StrictRedis):
    """Simple subclass that counts how many queries this connection ran.
    Also does some special handling of None and unicode objects.
    """
    def __init__(self, *args, **kwargs):
        self.trace = kwargs.pop('trace', False)
        super(TracingRedis, self).__init__(*args, **kwargs)
        self.log = logging.getLogger('redis.trace')

    def pipeline(self, context=None, transaction=True, shard_hint=None):
        self.connection_pool.total_pipelines += 1
        if self.trace:
            self.log.debug("[%04d] PIPELINE BEGIN",
                           self.connection_pool.total_commands)
        return TracingPipeline(self.connection_pool, self.response_callbacks,
                               transaction, shard_hint)

    def execute_command(self, *args, **kwargs):
        # self.connection_pool.total_commands += 1
        if self.trace:
            self.log.debug('[%04d] %s', self.connection_pool.total_commands,
                           args)
        return super(TracingRedis, self).execute_command(*args, **kwargs)


def mangler(void, value):
    """mangling function to replace `encode` in subclasses of a redis
    connection, responsible for delivering unicode as bytestrings, as well as
    handling null responses as ''
    """
    # encode unicode strings to utf-8, None -> ''
    if isinstance(value, unicode):
        return value.encode('utf8')
    elif value is None:
        return ''
    else:
        return str(value)


class ManglingConnection(redis.Connection):
    encode = mangler


class ManglingUnixDomainSocketConnection(redis.UnixDomainSocketConnection):
    encode = mangler


class CountingPool(redis.ConnectionPool):
    def __init__(self, *args, **kwargs):
        super(CountingPool, self).__init__(*args, **kwargs)
        self.reset_counts()
        self.log = logging.getLogger('redis.counting_pool')
        self.log.debug("Creating new pool")

    def reset_counts(self):
        self.total_connections_leased = 0
        self.total_commands = 0
        self.total_pipelines = 0
        self.total_pipeline_commands = 0

    def get_connection(self, *args, **kwargs):
        self.total_connections_leased += 1
        c = super(CountingPool, self).get_connection(*args, **kwargs)
        self.log.debug("Getting connection from pool (id: %s)", id(c))
        return c

    def make_connection(self, *args, **kwargs):
        # get_connection will call make_connection as needed, so we don't want
        # to count another leased connection here.
        c = super(CountingPool, self).make_connection(*args, **kwargs)
        self.log.debug("Making new connection (id: %s)", id(c))
        return c


class ConnectionManager(object):
    connections = {}

    def __init__(self, host=None, port=None, unix_socket_path=None, dbnum=0,
                 trace=False, decode_responses=True):
        """set module level configuration

        :param host: hostname for the redis server
        :param port: port for the redis server
        :param dbnum: database number to connect to
        :parma decode_responses: What to set the decode_responses option on the
               connection objects to.
        """
        self.log = logging.getLogger('redis')
        self.host = host
        self.port = port
        self.unix_socket_path = unix_socket_path
        self.dbnum = dbnum
        self.trace = trace
        self.decode_responses = decode_responses

        sane_config = False
        if self.unix_socket_path is not None:
            self.log.info("Initializing redis connection pool on UNIX:%s "
                          "[db:%s]", self.unix_socket_path, self.dbnum)
            sane_config = True
        elif self.host is not None and self.port is not None:
            self.log.info("Initializing redis connection pool on "
                          "TCP:%s:%s [db:%s]", self.host, self.port,
                          self.dbnum)
            sane_config = True
        else:
            self.log.error("Connection pool cannot be initialized unless "
                           "either host and port are set, or "
                           "unix_socket_path is set")

        if sane_config:
            self._get_pool()
        else:
            # replace our pool getter with nothing
            self._get_pool = lambda: None

    def init_app(self, app):
        """This callback can be used to initialize an application for the
        use with this database setup.  Never use a database in the context
        of an application not initialized that way or connections will
        leak.
        """
        self.app = app
        self.host = app.config.get('REDIS_HOST', 'localhost')
        self.port = app.config.get('REDIS_PORT', 6379)
        if app.config.get('TESTING'):
            self.dbnum = app.config.get('REDIS_TEST_DBNUM', 0)
        else:
            self.dbnum = app.config.get('REDIS_DBNUM', 0)
        self.decode_responses = app.config.get('REDIS_DECODE_RESPONSES', True)

    def __str__(self):
        if self.unix_socket_path:
            return "<ConnectionManager UNIX:%s [db:%s] %x>" % (
                self.unix_socket_path, self.dbnum, id(self))
        else:
            return "<ConnectionManager TCP:%s:%s [db:%s] %x>" % (
                self.host, self.port, self.dbnum, id(self))

    def _get_pool(self):
        """Create a connection pool using our custom connection type, and cache
        the result at the class level
        """
        cache_key = "%s:%s:%s:%s" % (self.unix_socket_path, self.host,
                                     self.port, self.dbnum)
        pool = ConnectionManager.connections.get(cache_key)
        self.log.debug('getting a pool for %s', cache_key)

        if pool is None:
            # hasn't been created yet, let's make it
            self.log.debug("creating new redis pool on %s", cache_key)
            pool_options = {
                'db': self.dbnum,
                'decode_responses': self.decode_responses,
                'encoding': 'utf-8'
            }
            if self.unix_socket_path:
                pool_options['connection_class'] = \
                    ManglingUnixDomainSocketConnection
                pool_options['path'] = self.unix_socket_path
            else:
                pool_options['connection_class'] = ManglingConnection
                pool_options['host'] = self.host
                pool_options['port'] = self.port

            pool = CountingPool(**pool_options)
            ConnectionManager.connections[cache_key] = pool
        return pool

    @classmethod
    def get_summary(klass):
        out = {}
        for key, pool in klass.connections.items():
            out[key] = {
                'key': key,
                'total_connections_leased': pool.total_connections_leased,
                'total_commands': pool.total_commands,
                'total_pipelines': pool.total_pipelines,
                'total_pipeline_commands': pool.total_pipeline_commands,
                'full_total': (pool.total_commands +
                               pool.total_pipeline_commands)
                }
        return out

    @classmethod
    def reset_counts(klass):
        for pool in klass.connections.values():
            pool.reset_counts()

    @contextmanager
    def connection(self, context=None):
        con = None
        try:
            con = TracingRedis(trace=self.trace,
                               connection_pool=self._get_pool())
            yield con
        except Exception:
            self.log.exception("Something blew up in the redis "
                               "context manager")
            raise
        finally:
            self.log.debug("returning connection to pool")
            if con:
                del con

    @contextmanager
    def pipeline(self, context=None):
        con = None
        try:
            con = TracingRedis(trace=self.trace,
                               connection_pool=self._get_pool())
            with con.pipeline() as pipe:
                try:
                    yield pipe
                    pipe.execute()
                except:
                    self.log.exception("Something blew up inside "
                                       "a pipeline context")
                    pipe.reset()
                    raise
        except Exception:
            self.log.exception("Something blew up in the redis "
                               "context manager")
            raise
        finally:
            self.log.debug("returning connection to pool")
            if con:
                del con

    def raw_connection(self):
        """Get a raw connection object from the pool, and return it to the
        client, this is most useful to third party libs or simple scripts that
        want direct access to a StrictRedis-like connection object

        You should close or del this object as soon as you're done with it.
        """
        return TracingRedis(trace=self.trace, connection_pool=self._get_pool())


class ShardedConnectionManager(object):
    @classmethod
    def from_config(klass, conf):
        # some consistency checks first...
        pool_map = {}
        shard_map = {}
        for i in xrange(conf.shard_count):
            shard_map[i] = None

        for server_info in conf.member_servers:
            # add this server to our pool map
            pool_map[server_info.url] = None

            # add all the shards this server is responsible for (ranges in
            # config are inclusive so add 1 to the end)
            shard_list = xrange(server_info.first_shard,
                                server_info.last_shard + 1)
            for shard in shard_list:
                # this shard should already be in the shard_map, but with a
                # None value
                if not shard_map in shard:
                    raise ValueError("Shard range found in member %s is out "
                                     "of bounds for the set!" % server_info)
                elif shard_map[shard] is not None:
                    raise ValueError("Shard %s is claimed by more than 1 "
                                     "server!" % shard)
                shard_map[shard] = server_info.url

        # at this point all shards should be claimed
        if not all(shard_map.values()):
            raise ValueError("Shard set claims there are %s shards, but not "
                             "all are claimed by a member server" %
                             conf.shard_count)

        scm = klass(shard_map, pool_map)
        scm.set_ready(True)
        return scm

    @staticmethod
    def parse_url(url):
        import urlparse
        u = urlparse.urlparse(url)
        return {
            'db': int(u.path.replace("/", "") or 0),
            'host': u.hostname,
            'port': int(u.port or 6379)
        }

    def __init__(self, shard_map, pool_map):
        self.log = logging.getLogger('redis_sharded')
        self.pool_map = pool_map
        self.shard_map = shard_map
        for url in pool_map.keys():
            # a list of redis urls that we need to make pools for
            self.log.debug("mapping member server: %s", url)
            pool = redis.connection.ConnectionPool(**self.parse_url(url))
            pool.url = url
            self.pool_map[url] = pool

        self.log.info("Shard Map created %d shards over %d servers",
                      len(shard_map), len(pool_map))
        self.__ready = False

    def set_ready(self, is_ready=True):
        self.__ready = is_ready

    def shard_node(self, node_id):
        """Consistently hash node_id, and figure out which shard the hash
        belongs to """
        # use crc32 (adler32 is faster, but has more collisions)
        return zlib.crc32("%d" % node_id) % len(self.shard_map.keys())

    def flushdb(self):
        """Destroy every shard's db
        """
        for pool in self.pool_map.values():
            con = StrictRedis.from_url(pool.url, connection_pool=pool)
            self.log.debug("flushing shard member: %s", con)
            con.flushdb()
            del con

    @contextmanager
    def connection(self, node_id):
        if not self.__ready:
            return
        shard = self.shard_node(node_id)
        try:
            pool = self.pool_map.get(self.shard_map.get(shard))
            con = StrictRedis.from_url(pool.url, connection_pool=pool)
            yield con
        except Exception:
            self.log.exception("Something blew up in the Redis context "
                               "manager")
            raise
        finally:
            del con

    @contextmanager
    def pipeline(self, node_id):
        if not self.__ready:
            return
        shard = self.shard_node(node_id)
        try:
            pool = self.pool_map.get(self.shard_map.get(shard))
            con = StrictRedis.from_url(pool.url, connection_pool=pool)
            with con.pipeline() as pipe:
                try:
                    yield pipe
                    pipe.execute()
                except:
                    self.log.exception("Something blew up inside a pipeline "
                                       "context")
                    pipe.reset()
                    raise
        except Exception:
            self.log.exception("Something blew up in the Redis context "
                               "manager")
            raise
        finally:
            del con
